--	Block.lua - Class representing a block.
--  ----------------------------------------------------------------------------
--	This file is part of 'EXCOSC Base' a simple OSC-aware game skeleton.
--	Copyright (C) 2013  Niall Moody
--	
--	This program is free software: you can redistribute it and/or modify
--	it under the terms of the GNU General Public License as published by
--	the Free Software Foundation, either version 3 of the License, or
--	(at your option) any later version.
--
--	This program is distributed in the hope that it will be useful,
--	but WITHOUT ANY WARRANTY; without even the implied warranty of
--	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--	GNU General Public License for more details.
--
--	You should have received a copy of the GNU General Public License
--	along with this program.  If not, see <http://www.gnu.org/licenses/>.
--	----------------------------------------------------------------------------

--	Used to give each block a unique id.
globalBlockId = 0

-- Class declaration.
class 'Block'

--------------------------------------------------------------------------------
--	The class constructor.
function Block:__init(pos, col)
	self.position = TwoFloats(pos)
	self.colour = ThreeFloats(col)
	self.newColour = ThreeFloats(col)
	self.colIndex = 0.0
	self.typeName = "Wall"
	self.blockId = globalBlockId
	globalBlockId = globalBlockId + 1

	self.neighbours = {}

	--Used by GameState.lua to determine when it needs to update the block's
	--neighbours.
	self.hasMoved = false

	--Used by GameState.lua to determine when to delete the block.
	self.canDelete = false
end

--------------------------------------------------------------------------------
--	Updates the block.
function Block:update(app)
	--Update the block's colour.
	if (self.colour ~= self.newColour) then
		self.colIndex = self.colIndex + 0.02

		if (self.colIndex >= 1.0) then
			self.colIndex = 0.0
			self.colour = self.newColour
		end
	else
		--Take on the colour of any neighbours whose colour has changed.
		for i=1,#self.neighbours do
			if (self.neighbours[i].colIndex > 0.5) then
				self.newColour = self.neighbours[i].newColour
			end
		end
	end
end

--------------------------------------------------------------------------------
--	Draws the block.
function Block:draw(d)
	tempCol = ThreeFloats(self.colour + ((self.newColour-self.colour) * self.colIndex))

	d:drawImage(gridImages[1],
				self.position-24.0,
				TwoFloats(1.0),
				tempCol,
				1.0)
end

--------------------------------------------------------------------------------
--	Called when the block is hit by a shockwave.
function Block:onShockwave(app)
	particles:trigger(self.position, self.colour)
	self.canDelete = true
end

--------------------------------------------------------------------------------
--	Called to move the block.
function Block:setPosition(pos)
	self.position = pos
	self.hasMoved = true
end

--------------------------------------------------------------------------------
--	Clears the block's neighbours table.
--	(and informs any neighbours that they are now missing a neighbour)
function Block:clearNeighbours()
	for i=#self.neighbours,1,-1 do
		for j=1,#self.neighbours[i].neighbours do
			if (self.neighbours[i].neighbours[j].blockId == self.blockId) then
				table.remove(self.neighbours[i].neighbours, j)

				break
			end
		end

		table.remove(self.neighbours, i)
	end
end
