--	Enemy.lua - The enemy class.
--  ----------------------------------------------------------------------------
--	This file is part of 'EXCOSC Base' a simple OSC-aware game skeleton.
--	Copyright (C) 2013  Niall Moody
--	
--	This program is free software: you can redistribute it and/or modify
--	it under the terms of the GNU General Public License as published by
--	the Free Software Foundation, either version 3 of the License, or
--	(at your option) any later version.
--
--	This program is distributed in the hope that it will be useful,
--	but WITHOUT ANY WARRANTY; without even the implied warranty of
--	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--	GNU General Public License for more details.
--
--	You should have received a copy of the GNU General Public License
--	along with this program.  If not, see <http://www.gnu.org/licenses/>.
--	----------------------------------------------------------------------------

-- Class declaration.
class 'Enemy'

--------------------------------------------------------------------------------
--	Handles collisions with walls.
function enemyWallCollisions(enemy, newPos)
	colourDistSq = 144.0 * 144.0
	retval = nil

	oldPos = TwoFloats(enemy.position)

	tempDir = TwoFloats(newPos - oldPos)
	topLeft = TwoFloats(newPos-20.0)
	bottomRight = TwoFloats(topLeft+40.0)

	enemy.position = newPos

	for i=0,#blocks do
		if (enemy.position:getDistanceSquared(blocks[i].position) < colourDistSq) then
			if (blocks[i].colour ~= enemy.colour) and (blocks[i].colIndex <= 0.0) then
				blocks[i].newColour = enemy.colour
			end

			if (blocks[i].typeName ~= "Water") and
			   (blocks[i].typeName ~= "Fire") and
			   (blocks[i].typeName ~= "Bomb") then
				blockTL = TwoFloats(blocks[i].position-22.0)
				blockBR = TwoFloats(blockTL+44.0)

				if ((topLeft.x < blockBR.x) and (bottomRight.x > blockTL.x)) and
				   ((topLeft.y < blockBR.y) and (bottomRight.y > blockTL.y)) then
				   --If the x-axis overlap is greater than the y-axis
					if (math.abs(enemy.position.x-blocks[i].position.x) > math.abs(enemy.position.y-blocks[i].position.y)) then
						--Overlapping on the left of the block.
						if (topLeft.x < blockTL.x) and (bottomRight.x > blockTL.x) then
							enemy.position.x = blocks[i].position.x-43.0
						--Overlapping on the right of the block.
						elseif (topLeft.x < blockBR.x) and (bottomRight.x > blockBR.x) then
							enemy.position.x = blocks[i].position.x+43.0
						end
					else
						--Overlapping on the top of the block.
						if (topLeft.y < blockTL.y) and (bottomRight.y > blockTL.y) then
							enemy.position.y = blocks[i].position.y-43.0
						--Overlapping on the bottom of the block.
						elseif (topLeft.y < blockBR.y) and (bottomRight.y > blockBR.y) then
							enemy.position.y = blocks[i].position.y+43.0
						end
					end

					--If the enemy's speeding, trigger any relevant blocks.
					if (enemy.speedExtra > 0.5) then
						if (blocks[i].typeName == "Expando") or
						   (blocks[i].typeName == "Expoding") or
						   (blocks[i].typeName == "Ice") or
						   (blocks[i].typeName == "Flammable") then
							blocks[i].shake = 0.0
						end
					end

					retval = blocks[i]
				end
			elseif (blocks[i].typeName == "Fire") then
				blockTL = TwoFloats(blocks[i].position-22.0)
				blockBR = TwoFloats(blockTL+44.0)

				if ((topLeft.x < blockBR.x) and (bottomRight.x > blockTL.x)) and
				   ((topLeft.y < blockBR.y) and (bottomRight.y > blockTL.y)) then
					enemy.canDelete = true
					particles:trigger(enemy.position, enemy.colour)
					particles:trigger(enemy.position, enemy.colour)

					retval = blocks[i]
				end
			end
		end
	end

	return retval
end

--------------------------------------------------------------------------------
--	Default random wander behaviour.
function behaviourWander(enemy)
	newPos = TwoFloats(enemy.position)

	--Move newPos along its direction vector.
	newPos = newPos + (enemy.direction * (enemy.speed + enemy.speedExtra))

	--Update direction.
	enemy.direction = enemy.direction + ((enemy.newDirection-enemy.direction)*0.1)
	if (enemy.direction:getDistance(enemy.newDirection) < 0.01) then
		enemy.direction = enemy.newDirection
		enemy.newDirection = TwoFloats((Random.getGlobal():getRandomFloat()*2.0)-1.0,
									   (Random.getGlobal():getRandomFloat()*2.0)-1.0):normal()
	end

	--Handle any block collisions.
	collidedBlock = enemyWallCollisions(enemy, newPos)
	if (collidedBlock ~= nil) then
		enemy.direction = (enemy.position-collidedBlock.position):normal()
		enemy.newDirection = TwoFloats((Random.getGlobal():getRandomFloat()*2.0)-1.0,
									   (Random.getGlobal():getRandomFloat()*2.0)-1.0)
	end
end

--------------------------------------------------------------------------------
--	Move directly towards the player behaviour.
function behaviourSeekPlayer(enemy)
	tempDir = (player.position - enemy.position):normal()

	newPos = TwoFloats(enemy.position + (tempDir * (enemy.speed + enemy.speedExtra) * 0.66))

	--Handle any block collisions.
	enemyWallCollisions(enemy, newPos)
end

--------------------------------------------------------------------------------
--	Speed right behaviour.
function behaviourSpeedRight(enemy)
	enemy.speedExtra = 2.0

	newPos = enemy.position
	newPos.x = newPos.x + (enemy.speed + enemy.speedExtra)

	--Handle any block collisions.
	enemyWallCollisions(enemy, newPos)

	if (enemy.behaviourTimer <= 0) then
		enemy.behaviour = behaviourWander
	end
end

--------------------------------------------------------------------------------
--	Speed left behaviour.
function behaviourSpeedLeft(enemy)
	enemy.speedExtra = 2.0

	newPos = enemy.position
	newPos.x = newPos.x - (enemy.speed + enemy.speedExtra)

	--Handle any block collisions.
	enemyWallCollisions(enemy, newPos)

	if (enemy.behaviourTimer <= 0) then
		enemy.behaviour = behaviourWander
	end
end

--------------------------------------------------------------------------------
--	Speed up behaviour.
function behaviourSpeedUp(enemy)
	enemy.speedExtra = 2.0

	newPos = enemy.position
	newPos.y = newPos.y - (enemy.speed + enemy.speedExtra)

	--Handle any block collisions.
	enemyWallCollisions(enemy, newPos)

	if (enemy.behaviourTimer <= 0) then
		enemy.behaviour = behaviourWander
	end
end

--------------------------------------------------------------------------------
--	Speed down behaviour.
function behaviourSpeedDown(enemy)
	enemy.speedExtra = 2.0

	newPos = enemy.position
	newPos.y = newPos.y + (enemy.speed + enemy.speedExtra)

	--Handle any block collisions.
	enemyWallCollisions(enemy, newPos)

	if (enemy.behaviourTimer <= 0) then
		enemy.behaviour = behaviourWander
	end
end

--------------------------------------------------------------------------------
--	Circle behaviour.
function behaviourCircle(enemy)
	newPos = TwoFloats(enemy.position)

	--Move newPos along its direction vector.
	newPos = newPos + (enemy.direction * (enemy.speed + enemy.speedExtra))

	--Move direction in a circle.
	enemy.posIndex = enemy.posIndex - 0.03
	if (enemy.posIndex < 0.0) then
		enemy.posIndex = enemy.posIndex + (2.0*math.pi)
	end

	enemy.direction.x = math.sin(enemy.posIndex)
	enemy.direction.y = math.cos(enemy.posIndex)

	--Handle any block collisions.
	collidedBlock = enemyWallCollisions(enemy, newPos)
	if (collidedBlock ~= nil) then
		enemy.direction = (enemy.position-collidedBlock.position):normal()
		enemy.newDirection = TwoFloats((Random.getGlobal():getRandomFloat()*2.0)-1.0,
									   (Random.getGlobal():getRandomFloat()*2.0)-1.0)
	end
end

--------------------------------------------------------------------------------
--	Bomb behaviour.
function behaviourBomb(enemy)
	--The enemy's movement stays the same.
	behaviourWander(enemy)

	--Update posIndex, drop a bomb when it passes 1.0.
	enemy.posIndex = enemy.posIndex + (0.005 + (Random.getGlobal():getRandomFloat() * 0.01))
	if (enemy.posIndex >= 1.0) then
		enemy.posIndex = 0.0

		addBomb(enemy.position, enemy.colour)
	end
end

--------------------------------------------------------------------------------
--	Flame behaviour.
function behaviourFlame(enemy)
	--The enemy's movement stays the same.
	behaviourWander(enemy)

	--Update posIndex, drop a bomb when it passes 1.0.
	enemy.posIndex = enemy.posIndex + (0.005 + (Random.getGlobal():getRandomFloat() * 0.01))
	if (enemy.posIndex >= 1.0) then
		enemy.posIndex = 0.0

		--We cheat a bit to try and make sure the enemy doesn't instantly kill
		--itself.
		addFire(enemy.position - (enemy.direction * 48.0))
	end
end

--------------------------------------------------------------------------------
--	Water behaviour.
function behaviourWater(enemy)
	--The enemy's movement stays the same.
	behaviourWander(enemy)

	--Update posIndex, drop a bomb when it passes 1.0.
	enemy.posIndex = enemy.posIndex + (0.01 + (Random.getGlobal():getRandomFloat() * 0.02))
	if (enemy.posIndex >= 1.0) then
		enemy.posIndex = 0.0

		addWater(enemy.position)
	end
end

--------------------------------------------------------------------------------
--	The class constructor.
function Enemy:__init(pos, col)
	self.position = TwoFloats(pos)
	self.colour = ThreeFloats(col)
	self.speed = 2.0
	self.speedExtra = 0.0
	self.behaviour = behaviourWander
	self.canDelete = false

	self.direction = TwoFloats((Random.getGlobal():getRandomFloat()*2.0)-1.0,
							   (Random.getGlobal():getRandomFloat()*2.0)-1.0):normal()
	self.newDirection = TwoFloats((Random.getGlobal():getRandomFloat()*2.0)-1.0,
								  (Random.getGlobal():getRandomFloat()*2.0)-1.0):normal()

	--Used by circle behaviour.
	self.posIndex = 0.0

	--Used to draw ghosts behind the enemy when it's speeding.
	self.ghosts = {TwoFloats(self.position),
				   TwoFloats(self.position),
				   TwoFloats(self.position),
				   TwoFloats(self.position),
				   TwoFloats(self.position),
				   TwoFloats(self.position),
				   TwoFloats(self.position),
				   TwoFloats(self.position)}

	--Timer used to ensure enemies aren't activated straight away when the
	--level begins.
	self.timer = 1.0

	--Used to ensure enemy behaviours don't get toggled too fast.
	self.behaviourTimer = 0
end

--------------------------------------------------------------------------------
--	Moves the enemy.
function Enemy:update(app)
	if (self.speedExtra > 0.0) then
		for i=#self.ghosts,2,-1 do
			self.ghosts[i] = TwoFloats(self.ghosts[i-1])
		end
		self.ghosts[1] = TwoFloats(self.position)
	end

	if (self.timer > 0.0) then
		self.timer = self.timer - 0.02
	end

	if (self.behaviourTimer > 0) then
		self.behaviourTimer = self.behaviourTimer - 1
	end

	self.behaviour(self)

	if (self.speedExtra > 0.0) then
		self.speedExtra = self.speedExtra - 0.05
		if (self.speedExtra < 0.0) then
			self.speedExtra = 0.0
		end
	end
end

--------------------------------------------------------------------------------
--	Draws the enemy.
function Enemy:draw(d)
	if (self.speedExtra > 0.5) then
		alpha = 1.0
		if (self.speedExtra < 1.0) then
			alpha = (self.speedExtra-0.5) * 2.0
		end

		for i=1,#self.ghosts do
			d:drawImage("Enemy",
						self.ghosts[i]-24.0,
						TwoFloats(1.0),
						self.colour,
						((1.0 - (i/#self.ghosts)) * alpha))
		end
	end

	d:drawImage("Enemy",
				self.position-24.0,
				TwoFloats(1.0),
				self.colour,
				1.0)
end

--------------------------------------------------------------------------------
--	Called to set the enemy's position via OSC.
function Enemy:setOscPositionX(pos)
	if (self.behaviourTimer <= 0) then
		self.position.x = pos
	end
end

--------------------------------------------------------------------------------
--	Called to set the enemy's position via OSC.
function Enemy:setOscPositionY(pos)
	if (self.behaviourTimer <= 0) then
		self.position.y = pos
	end
end
