--	ParticleGenerator.lua - A simple particle generator class.
--  ----------------------------------------------------------------------------
--	This file is part of 'EXCOSC Base' a simple OSC-aware game skeleton.
--	Copyright (C) 2013  Niall Moody
--	
--	This program is free software: you can redistribute it and/or modify
--	it under the terms of the GNU General Public License as published by
--	the Free Software Foundation, either version 3 of the License, or
--	(at your option) any later version.
--
--	This program is distributed in the hope that it will be useful,
--	but WITHOUT ANY WARRANTY; without even the implied warranty of
--	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--	GNU General Public License for more details.
--
--	You should have received a copy of the GNU General Public License
--	along with this program.  If not, see <http://www.gnu.org/licenses/>.
--	----------------------------------------------------------------------------

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
class 'Particle'

--------------------------------------------------------------------------------
function Particle:__init(pos, col)
	self.position = TwoFloats(pos)
	self.dir = TwoFloats((Random.getGlobal():getRandomFloat()*2.0)-1.0,
						 (Random.getGlobal():getRandomFloat()*2.0)-1.0)
	self.colour = ThreeFloats(col)
	self.life = 1.0
	self.decay = 0.005 + (Random.getGlobal():getRandomFloat()*0.045)
	self.gravity = TwoFloats()

	self.wave = 0.0
	self.waveIndex = 0.0
end

--------------------------------------------------------------------------------
function Particle:update()
	if ((self.gravity.x ~= 0.0) or (self.gravity.y ~= 0.0)) then
		self.dir = self.dir + ((self.gravity - self.dir) * 0.05)
	end

	self.position = self.position + (self.dir * 4.0 * self.life)
	if (self.wave > 0.0) then
		self.waveIndex = self.waveIndex + 0.1
		if (self.waveIndex > (2.0 * math.pi)) then
			self.waveIndex = 0.0
		end
	end

	self.life = self.life - self.decay
end

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
class 'ParticleGenerator'

--------------------------------------------------------------------------------
function ParticleGenerator:__init()
	self.particles = {}
end

--------------------------------------------------------------------------------
function ParticleGenerator:update()
	for i=#self.particles,1,-1 do
		self.particles[i]:update()

		if (self.particles[i].life <= 0.0) then
			table.remove(self.particles, i)
		end
	end
end

--------------------------------------------------------------------------------
function ParticleGenerator:draw(d)
	for i=1,#self.particles do
		tempLife = fakeLog(self.particles[i].life)
		tempPos = TwoFloats(self.particles[i].position-(8.0*tempLife))

		if (self.particles[i].wave > 0.0) then
			tempPerp = TwoFloats(self.particles[i].dir:perpendicular())
			tempPos = tempPos + (tempPerp * math.sin(self.particles[i].waveIndex) * self.particles[i].wave)
		end

		d:drawImage("Particle",
					tempPos,
					TwoFloats(tempLife),
					self.particles[i].colour,
					tempLife)
	end
end

--------------------------------------------------------------------------------
function ParticleGenerator:trigger(pos, col)
	numParticles = 8 + Random.getGlobal():getRandomInt(6)

	for i=0,numParticles do
		tempCol = ThreeFloats(col)
		tempCol = tempCol * (0.6 + (Random.getGlobal():getRandomFloat() * 0.4))
		self.particles[#self.particles+1] = Particle(pos, tempCol)
	end
end

--------------------------------------------------------------------------------
function ParticleGenerator:triggerSmoke(pos, col)
	numParticles = 12 + Random.getGlobal():getRandomInt(12)

	for i=0,numParticles do
		tempCol = ThreeFloats(col)
		tempCol = tempCol * (0.6 + (Random.getGlobal():getRandomFloat() * 0.4))
		self.particles[#self.particles+1] = Particle(pos, tempCol)
		self.particles[#self.particles].wave = 8.0
		self.particles[#self.particles].dir = TwoFloats((Random.getGlobal():getRandomFloat()*0.2)-0.1, -0.5)
	end
end

--------------------------------------------------------------------------------
function ParticleGenerator:triggerWater(pos, col)
	numParticles = 12 + Random.getGlobal():getRandomInt(12)

	for i=0,numParticles do
		tempCol = ThreeFloats(col)
		tempCol = tempCol * (0.6 + (Random.getGlobal():getRandomFloat() * 0.4))
		self.particles[#self.particles+1] = Particle(pos, tempCol)
		self.particles[#self.particles].dir = self.particles[#self.particles].dir * 0.5
		self.particles[#self.particles].gravity = TwoFloats(0.0, 0.5)
		self.life = 0.66
	end
end
