
--	GameState.lua - Main script run by LuaGameState.cpp.
--  ----------------------------------------------------------------------------
--	This file is part of 'EXCOSC Base' a simple OSC-aware game skeleton.
--	Copyright (C) 2013  Niall Moody
--	
--	This program is free software: you can redistribute it and/or modify
--	it under the terms of the GNU General Public License as published by
--	the Free Software Foundation, either version 3 of the License, or
--	(at your option) any later version.
--
--	This program is distributed in the hope that it will be useful,
--	but WITHOUT ANY WARRANTY; without even the implied warranty of
--	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--	GNU General Public License for more details.
--
--	You should have received a copy of the GNU General Public License
--	along with this program.  If not, see <http://www.gnu.org/licenses/>.
--	----------------------------------------------------------------------------

--	The size of the grid's tiles.
tileSize = 48.0
--	The size of the grid's x-axis.
gridX = 0
--	The size of the grid's y-axis.
gridY = 0
--	The grid offset.
gridOffset = TwoFloats()

--	The grid itself.
--	0 == space
--	1 == basic wall
--	2 == fire
--	3 == water
--	4 == ice
--	5 == flammable
--	6 == exploding
--	7 == expando
--	8 == bomb
--	100 == corridor (during level generation only)
--	101 == room (during level generation only)
grid = {}

--	The grid images.
gridImages = {"BlockBase",
			  "Fire",
			  "Water",
			  "IceBlock",
			  "FlammableBlock",
			  "ExplodingBlock",
			  "ExpandoBlock",
			  "Bomb"}

--	The block objects that make up the map.
blocks = {}

--	The player object.
player = Player()
--	The position of the stairs.
stairsPos = TwoFloats()
--	Used to emphasise the player at the start of a level.
playerEmphasis = 0.0

--	The second player's position.
player2Pos = TwoFloats(-100)
--	The second player's colour.
player2Col = ThreeFloats(1.0, 0.5, 0.5)

--	The active enemies.
enemies = {}
--	The enemy behaviour pads.
pads = {}

--	Any active shockwave objects.
shockwaves = {}
--	Our particle generator.
particles = ParticleGenerator()

--	The background colour.
backColour = hslToRgb(ThreeFloats(Random.getGlobal():getRandomFloat(), 0.5, 0.6))
--	The next background colour.
nextBack = hslToRgb(ThreeFloats(Random.getGlobal():getRandomFloat(), 0.5, 0.6))
--	Used to interpolate the background colour.
backInterp = 0.0

--	Which level the player is currently on.
level = 1
--	The current high score.
hiScore = 1

--	Displays the title text if > 0.
title = 1.0
--	The title layout.
titleLayout = TextLayout()
--	The subtitle layout.
subtitleLayout = TextLayout()

--	Queue for any sounds to be played.
soundsQueue = {}

--------------------------------------------------------------------------------
--	Called when the state is first initialised.
function initialise(app)
	gridX = math.floor(app:getDrawingAreaSize().x/tileSize)
	gridY = math.floor((app:getDrawingAreaSize().y-64.0)/tileSize)
	gridOffset = TwoFloats((app:getDrawingAreaSize().x-(gridX*tileSize))/2.0,
						   64.0 + ((app:getDrawingAreaSize().y-64.0-(gridY*tileSize))/2.0))

	for j=0,(gridY-1) do
		grid[j] = {}
	
		for i=0,(gridX-1) do
			grid[j][i] = 1
		end
	end

	--Generate the level.
	generateLevel()
	createBlocks()

	--Set the player's starting position.
	player.position = TwoFloats(gridOffset + (getPlayerStart() * 48.0) + 24.0)
	--Set the stairs' position.
	stairsPos = TwoFloats(gridOffset + (getStairsPos() * 48.0) + 24.0)

	--Add the enemies.
	addEnemies()
	--Add the enemy behaviour pads.
	addPads()

	--Fill out all the blocks' neighbours tables.
	for i=0,#blocks do
		fillBlockNeighbours(i, false)
	end

	--Turn on audio fx.
	app:toggleDelay(true)
	app:setDelayParams(0.33, 0.66, 0.33)
	--app:toggleGranulator(true)
	--app:setGranulatorParams(1.0, 0.5, 2.0, 0.0, 0.33, 12.0)
	--app:toggleDistortion(true)
	--app:setDistortionGain(32.0)
end

--------------------------------------------------------------------------------
--	Used to update game logic etc.
function update(app)
	if (title < 1.0) then
		--Update the player.
		player:update(app)

		--Update the enemies.
		for i=1,#enemies do
			enemies[i]:update(app)
		end

		--Delete any expired enemies.
		for i=#enemies,1,-1 do
			if enemies[i].canDelete then
				table.remove(enemies, i)
			end
		end

		enemyCollisions()

		--Update shockwaves.
		for i=1,#shockwaves do
			shockwaves[i]:update(app)
		end

		--Delete any expired shockwaves.
		for i=#shockwaves,1,-1 do
			if (shockwaves[i].radius >= 1.0) then
				table.remove(shockwaves, i)
			end
		end

		--Update blocks.
		for i=#blocks,0,-1 do
			blocks[i]:update(app)
		end

		--Delete any blocks that should be deleted.
		for i=#blocks,0,-1 do
			if blocks[i].canDelete then
				blocks[i]:clearNeighbours()
				table.remove(blocks, i)
			end
		end

		--Update neighbours tables for any blocks which have moved.
		for i=0,#blocks do
			if blocks[i].hasMoved then
				blocks[i]:clearNeighbours()
				fillBlockNeighbours(i, true)
				blocks[i].hasMoved = false
			end
		end

		--Update pads.
		for i=1,#pads do
			pads[i]:update(app)
		end

		--Update particle generator.
		particles:update()

		--Test if the player's hit the stairs.
		if (player.position:getDistance(stairsPos) < 36.0) then
			setupNewLevel()
			level = level + 1
			if (level > hiScore) then
				hiScore = level
			end

			soundsQueue[#soundsQueue+1] = "GoDownStairs"
		end

		--If we need to emphasise the player.
		if (playerEmphasis < (math.pi * 2.0)) then
			tempPos = TwoFloats(player.position)
			tempPos.x = tempPos.x + (math.sin(playerEmphasis) * 64.0)
			tempPos.y = tempPos.y + (math.cos(playerEmphasis) * 64.0)

			particles:trigger(tempPos, player.colour)

			playerEmphasis = playerEmphasis + 0.33
		end

		--Test if the player's dead.
		if (player.hearts <= 0) then
			player.hearts = 3
			setupNewLevel()
			level = 1
			title = 3.0
		end

		--Send OSC messages for all the active enemies.
		for i=1,#enemies do
			app:sendOscMessage("/entity" .. i-1 .. "/position/x",
							   enemies[i].position.x/app:getDrawingAreaSize().x)
			app:sendOscMessage("/entity" .. i-1 .. "/position/y",
							   enemies[i].position.y/app:getDrawingAreaSize().y)
			app:sendOscMessage("/entity" .. i-1 .. "/colour/r",
							   enemies[i].colour.x)
			app:sendOscMessage("/entity" .. i-1 .. "/colour/g",
							   enemies[i].colour.y)
			app:sendOscMessage("/entity" .. i-1 .. "/colour/b",
							   enemies[i].colour.z)
		end

		--And the player.
		app:sendOscMessage("/player/position/x",
						   player.position.x/app:getDrawingAreaSize().x)
		app:sendOscMessage("/player/position/y",
						   player.position.y/app:getDrawingAreaSize().y)
		app:sendOscMessage("/player/colour/r", player.colour.x)
		app:sendOscMessage("/player/colour/g", player.colour.y)
		app:sendOscMessage("/player/colour/b", player.colour.z)
	end

	--Update the background colour.
	backInterp = backInterp + 0.005
	if (backInterp >= 1.0) then
		backColour = ThreeFloats(nextBack)
		nextBack = hslToRgb(ThreeFloats(Random.getGlobal():getRandomFloat(), 0.5, 0.6))

		backInterp = backInterp - 1.0
	end

	--Update the title alpha if necessary.
	if (title < 1.0) and (title > 0.0) then
		title = title - 0.02
	end

	--Play any queued sounds.
	for i=#soundsQueue,1,-2 do
		app:triggerSound(soundsQueue[#soundsQueue], 2.0, 0.5)
		table.remove(soundsQueue, i)
	end
end

--------------------------------------------------------------------------------
--	Draws the game.
function draw(d)
	d:fillRect(TwoFloats(0.0),
			   d:getDrawingAreaSize(),
			   backColour + ((nextBack-backColour) * backInterp),
			   1.0)

	d:drawTiledImage("Floor",
					 gridOffset,
					 TwoFloats(gridX*48.0, gridY*48.0),
					 TwoFloats(0.0),
					 TwoFloats(gridX, gridY),
					 backColour + ((nextBack-backColour) * backInterp),
					 1.0)

	d:drawImage("Stairs",
				stairsPos-24.0,
				TwoFloats(1.0),
				player.colour,
				1.0)

	for i=1,#pads do
		pads[i]:draw(d)
	end

	for i=0,#blocks do
		blocks[i]:draw(d)
	end

	for i=1,#enemies do
		enemies[i]:draw(d)
	end

	player:draw(d)

	for i=1,#shockwaves do
		shockwaves[i]:draw(d)
	end

	particles:draw(d)

	d:drawText("Level: " .. level,
			   "MedFont",
			   TwoFloats(32.0, 68.0),
			   256.0,
			   0,
			   1.0,
			   ThreeFloats(1.0),
			   1.0)

	d:drawText("High Score: " .. hiScore,
			   "MedFont",
			   TwoFloats(0.0, 68.0),
			   d:getDrawingAreaSize().x,
			   1,
			   1.0,
			   ThreeFloats(1.0),
			   1.0)

	if (player.hearts > 0) then
		d:drawImage("Heart",
					TwoFloats(d:getDrawingAreaSize().x-72.0-12.0, 8.0),
					TwoFloats(1.0),
					ThreeFloats(hslToRgb(ThreeFloats(0.0, 0.7, 0.8))),
					1.0)
	end
	if (player.hearts > 1) then
		d:drawImage("Heart",
					TwoFloats(d:getDrawingAreaSize().x-136.0-12.0, 8.0),
					TwoFloats(1.0),
					ThreeFloats(hslToRgb(ThreeFloats(0.0, 0.7, 0.8))),
					1.0)
	end
	if (player.hearts > 2) then
		d:drawImage("Heart",
					TwoFloats(d:getDrawingAreaSize().x-200.0-12.0, 8.0),
					TwoFloats(1.0),
					ThreeFloats(hslToRgb(ThreeFloats(0.0, 0.7, 0.8))),
					1.0)
	end

	--Draw title text if necessary.
	if (title > 0.0) then
		alpha = title
		if (alpha > 1.0) then
			alpha = 1.0
		end

		tempCol = ThreeFloats(backColour + ((nextBack-backColour) * backInterp))
		tempCol = ThreeFloats(tempCol.y, tempCol.z, tempCol.x) * 1.5
		d:drawImage("TitleText",
					TwoFloats((d:getDrawingAreaSize().x*0.5)-512.0, 0.0),
					TwoFloats(1.0),
					tempCol,
					alpha)
	end
end

--------------------------------------------------------------------------------
--	Called when the user depresses a key or a joystick button.
function keyDown(app, action, keyPressed, keyModifier)
	if (title < 1.0) then
		if (keyPressed == "z") or (keyPressed == "/") or (keyPressed == "Joy0Button0") then
			addBomb(player.position, player.colour)
			app:sendOscMessage("/player/trigger", 1.0)
		else
			player:keyDown(keyPressed)
		end
	else
		if (title > 1.0) then
			title = 1.0
		end
		title = title - 0.01
	end
end

--------------------------------------------------------------------------------
--	Called when the user releases a key or a joystick button.
function keyUp(app, action, keyPressed, keyModifier)
	if (keyPressed == "z") then
	
	else
		player:keyUp(keyPressed)
	end
end

--------------------------------------------------------------------------------
--	Called when the user depresses a mouse button.
function mouseDown(app, button, pos)
	
end

--------------------------------------------------------------------------------
--	Called when the user releases a mouse button.
function mouseUp(app, button, pos)

end

--------------------------------------------------------------------------------
--	Called when the user moves the mouse.
function mouseMove(app, pos)
	
end

--------------------------------------------------------------------------------
--	Called when the user moves the mousewheel.
function mouseWheel(app, dir)

end

--------------------------------------------------------------------------------
--	Called when the user moves a joystick.
function joyMove(app, action, value, joystick, axis)
	if (joystick == 0) then
		player:joyMove(axis, value)
	end

	if (title >= 1.0) then
		title = title - 0.01
	end
end

--------------------------------------------------------------------------------
--	Called when we receive an OSC message.
function oscMessageReceived(app, address, value)
	if (title < 1.0) then
		--Obtain the 1st 7 characters of the OSC address to check which object it's
		--meant for.
		rootAddr = address:sub(1, 7)

		if (rootAddr == "/player") then
			paramAddr = address:sub(8)

			if (paramAddr == "/position/x") then
				player2Pos.x = value * app:getDrawingAreaSize().x
			elseif (paramAddr == "/position/y") then
				player2Pos.y = value * app:getDrawingAreaSize().y
			elseif (paramAddr == "/colour/r") then
				player2Col.x = value
			elseif (paramAddr == "/colour/g") then
				player2Col.y = value
			elseif (paramAddr == "/colour/b") then
				player2Col.z = value
			elseif (paramAddr == "/trigger") then
				addBomb(player2Pos, player2Col)
			end
		elseif (rootAddr == "/entity") then
			--Entities are numbered, so we need to extract the entity index from its
			--address.
			index = tonumber(address:sub(8, 8))
			paramAddr = address:sub(9)

			if ((index+1) <= #enemies) then
				if (paramAddr == "/position/x") then
					enemies[index+1]:setOscPositionX(value * app:getDrawingAreaSize().x)
				elseif (paramAddr == "/position/y") then
					enemies[index+1]:setOscPositionY(value*app:getDrawingAreaSize().x)
				elseif (paramAddr == "/colour/r") then
					enemies[index+1].colour.x = value
				elseif (paramAddr == "/colour/g") then
					enemies[index+1].colour.y = value
				elseif (paramAddr == "/colour/b") then
					enemies[index+1].colour.z = value
				end
			end
		end
	end
end

--------------------------------------------------------------------------------
--	Sets up a new level.
function setupNewLevel()
	--Get rid of existing blocks.
	for i=#blocks,0,-1 do
		table.remove(blocks, i)
	end

	--Get rid of existing shockwaves.
	for i=#shockwaves,1,-1 do
		table.remove(shockwaves, i)
	end

	--Clear the grid.
	for j=0,(gridY-1) do
		for i=0,(gridX-1) do
			grid[j][i] = 1
		end
	end

	--Generate the level.
	generateLevel()
	createBlocks()

	--Set the player's starting position.
	player.position = TwoFloats(gridOffset + (getPlayerStart() * 48.0) + 24.0)
	--Set the stairs' position.
	stairsPos = TwoFloats(gridOffset + (getStairsPos() * 48.0) + 24.0)

	--Add the enemies.
	addEnemies()
	--Add the enemy behaviour pads.
	addPads()

	--Fill out all the blocks' neighbours tables.
	for i=0,#blocks do
		fillBlockNeighbours(i, false)
	end

	playerEmphasis = 0.0
end

--------------------------------------------------------------------------------
--	Handles collisions between enemies, the player, and the behaviour pads.
function enemyCollisions()
	for i=1,#enemies do
		distSq = enemies[i].position:getDistanceSquared(player.position)
		minDist = 44.0 * 44.0

		if (distSq < minDist) then
			hearts = player.hearts
			player:hurt(enemies[i].colour)

			if (hearts ~= player.hearts) then
				dir = TwoFloats((player.position - enemies[i].position):normal())
				player.position = enemies[i].position + (dir * 44.0)

				--soundsQueue[#soundsQueue+1] = "LoseHeart"
			end
		end

		--if (enemies[i].behaviourTimer <= 0) then
			for j=1,#pads do
				if (pads[j].timer <= 0) then
					distSq = enemies[i].position:getDistanceSquared(pads[j].position)
					minDist = 40.0 * 40.0

					if (distSq < minDist) then
						enemies[i].behaviour = pads[j].behaviour
						enemies[i].colour = pads[j].colour
						enemies[i].behaviourTimer = 60
						pads[j].timer = 60

						particles:trigger(enemies[i].position, pads[j].colour)

						if (pads[j].behaviour == behaviourSpeedRight) or
						   (pads[j].behaviour == behaviourSpeedRight) or
						   (pads[j].behaviour == behaviourSpeedRight) or
						   (pads[j].behaviour == behaviourSpeedRight) then
							soundsQueue[#soundsQueue+1] = "SpeedPadTriggered"
						else
							soundsQueue[#soundsQueue+1] = "PadTriggered"
						end
					end
				end
			end
		--end
	end
end

--------------------------------------------------------------------------------
--	Fills out the neighbours table for the indexed block.
--	index == The index of the block whose table to update.
--	updateNeighboursToo == If true, add the indexed block to all it's
--						   neighbours' neighbours table.
function fillBlockNeighbours(index, updateNeighboursToo)
	for i=0,#blocks do
		if (blocks[index].position:getDistance(blocks[i].position) < 50.0) then
			blocks[index].neighbours[#blocks[index].neighbours+1] = blocks[i]

			if updateNeighboursToo then
				blocks[i].neighbours[#blocks[i].neighbours+1] = blocks[index]
			end
		end
	end
end

--------------------------------------------------------------------------------
--	Culls any blocks which overlap.
function cullOverlapping(pos, id)
	minDistSq = 4.0 * 4.0

	for i=#blocks,0,-1 do
		if (pos:getDistanceSquared(blocks[i].position) < minDistSq) then
			if (blocks[i].blockId ~= id) then
				blocks[i]:clearNeighbours()
				table.remove(blocks, i)
			end
		end
	end
end

--------------------------------------------------------------------------------
--	Culls any blocks which overlap (special Expando version).
function cullOverlappingExpando(pos, id, index)
	minDistSq = 4.0 * 4.0

	for i=#blocks,0,-1 do
		if (pos:getDistanceSquared(blocks[i].position) < minDistSq) then
			if (blocks[i].blockId ~= id) then
				if (blocks[i].typeName == "Exploding") then
					blocks[i].shake = 0.0
					blocks[index]:clearNeighbours()
					table.remove(blocks, index)
					break
				elseif (blocks[i].typeName == "Water") then
					blocks[index]:clearNeighbours()
					table.remove(blocks, index)
					break
				elseif (blocks[i].typeName == "Expando") then
					if (index <= #blocks) then
						--Replace the Expando block with a normal block.
						blocks[#blocks+1] = Block(blocks[i].position, blocks[index].colour)

						blocks[index]:clearNeighbours()
						table.remove(blocks, index)

						blocks[i]:clearNeighbours()
						table.remove(blocks, i)
					end
					break
				else
					blocks[i]:clearNeighbours()
					table.remove(blocks, i)
				end
			end
		end
	end
end

--------------------------------------------------------------------------------
--	Adds a water block.
function addWater(pos)
	tempPos = TwoFloats(pos)

	blocks[#blocks+1] = Water(tempPos, ThreeFloats(1.0))
	fillBlockNeighbours(#blocks, true)
	cullOverlapping(pos, blocks[#blocks].blockId)

	soundsQueue[#soundsQueue+1] = "IceMelts"
end

--------------------------------------------------------------------------------
--	Adds 5 Waters when an Ice block melts.
function addIcePuddles(pos)
	tempPos = TwoFloats(pos)

	blocks[#blocks+1] = Water(tempPos, ThreeFloats(1.0))
	fillBlockNeighbours(#blocks, true)

	tempPos.x = tempPos.x - 48.0
	blocks[#blocks+1] = Water(tempPos, ThreeFloats(1.0))
	fillBlockNeighbours(#blocks, true)
	cullOverlapping(tempPos, blocks[#blocks].blockId)

	tempPos.x = tempPos.x + 96.0
	blocks[#blocks+1] = Water(tempPos, ThreeFloats(1.0))
	fillBlockNeighbours(#blocks, true)
	cullOverlapping(tempPos, blocks[#blocks].blockId)

	tempPos.x = tempPos.x - 48.0
	tempPos.y = tempPos.y - 48.0
	blocks[#blocks+1] = Water(tempPos, ThreeFloats(1.0))
	fillBlockNeighbours(#blocks, true)
	cullOverlapping(tempPos, blocks[#blocks].blockId)

	tempPos.y = tempPos.y + 96.0
	blocks[#blocks+1] = Water(tempPos, ThreeFloats(1.0))
	fillBlockNeighbours(#blocks, true)
	cullOverlapping(tempPos, blocks[#blocks].blockId)

	soundsQueue[#soundsQueue+1] = "IceMelts"
end

--------------------------------------------------------------------------------
--	Adds a Fire when a Flammable block is set on fire.
function addFire(pos)
	blocks[#blocks+1] = Fire(pos, ThreeFloats(1.0))
	fillBlockNeighbours(#blocks, true)
	cullOverlapping(pos, blocks[#blocks].blockId)
end

--------------------------------------------------------------------------------
--	Adds a Shockwave to the game.
function addShockwave(pos, col)
	shockwaves[#shockwaves+1] = Shockwave(pos, col)
	particles:trigger(pos, col)

	soundsQueue[#soundsQueue+1] = "BombExplodes"
end

--------------------------------------------------------------------------------
--	Adds 4 Expandos and 1 Wall when an Expando block expands.
function addExpandoBlocks(pos, col)
	tempPos = TwoFloats(pos)

	blocks[#blocks+1] = Block(tempPos, ThreeFloats(col))
	fillBlockNeighbours(#blocks, true)

	if (Random.getGlobal():getRandomFloat() < 0.66) then
		tempPos.x = tempPos.x - 48.0
		blocks[#blocks+1] = Expando(tempPos, ThreeFloats(col))
		fillBlockNeighbours(#blocks, true)
		cullOverlappingExpando(tempPos, blocks[#blocks].blockId, #blocks)
	end

	if (Random.getGlobal():getRandomFloat() < 0.66) then
		tempPos.x = tempPos.x + 96.0
		blocks[#blocks+1] = Expando(tempPos, ThreeFloats(col))
		fillBlockNeighbours(#blocks, true)
		cullOverlappingExpando(tempPos, blocks[#blocks].blockId, #blocks)
	end

	if (Random.getGlobal():getRandomFloat() < 0.66) then
		tempPos.x = tempPos.x - 48.0
		tempPos.y = tempPos.y - 48.0
		blocks[#blocks+1] = Expando(tempPos, ThreeFloats(col))
		fillBlockNeighbours(#blocks, true)
		cullOverlappingExpando(tempPos, blocks[#blocks].blockId, #blocks)
	end

	if (Random.getGlobal():getRandomFloat() < 0.66) then
		tempPos.y = tempPos.y + 96.0
		blocks[#blocks+1] = Expando(tempPos, ThreeFloats(col))
		fillBlockNeighbours(#blocks, true)
		cullOverlappingExpando(tempPos, blocks[#blocks].blockId, #blocks)
	end

	soundsQueue[#soundsQueue+1] = "ExpandoExpands"
end

--------------------------------------------------------------------------------
--	Adds a Bomb.
function addBomb(pos, col)
	blocks[#blocks+1] = Bomb(pos, col)
	fillBlockNeighbours(#blocks, true)

	soundsQueue[#soundsQueue+1] = "DropBomb"
end
